/**
 * $Id: Console.java,v 1.8 2009/07/28 22:55:08 juanca Exp $
 * editors.* is part of ACodeEditor
 * Copyright (C) 2009 Juan Carlos Rodrguez-del-Pino. All rights reserved.
 * license GNU/GPL, see LICENSE.txt or http://www.gnu.org/licenses/gpl-2.0.html
 * @author Juan Carlos Rodriguez-del-Pino
 **/
package editors;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.nio.charset.Charset;
import javax.swing.JApplet;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.WindowConstants;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

import util.I18n;

class ConsoleDoc extends DefaultStyledDocument implements Runnable{
	private static final long serialVersionUID = -852166104864394680L;
	public static final char CR='\r';
	public static final char LF='\n';
	public static final char TAB='\t';
	public static final char BS='\010';
	private final int capacity=10000;
	private MutableAttributeSet regularText, inputText;
	private Color background, foreground;
	private Socket connection=null;
	private OutputStream out = null;
	private InputStream in = null;
	private Highlights fonts=null;
	private JTextPane pane=null;
	private java.awt.Frame rootFrame=null;
	private String initialTitle=null;
	private StringBuffer textTypewrited=null;
	private Charset charset=null;
	private static Integer globalIdentification=0; //Console identification number 
	private int identification=0; //Console identification number 
	private boolean isDischargeable(char c){
		if(c >= ' ') return false;
		if(c == CR ) return false;
		if(c == LF ) return false;
		if(c == TAB ) return false;
		return true;
	}
    private void clear(){
    	try {
			super.remove(0, getLength());
		} catch (BadLocationException e) {}   	
    }
    private void append(String s, javax.swing.text.AttributeSet a){
    	try {
			if(getLength()>capacity)
				super.remove(0, getLength()-capacity);
			super.insertString(getLength(), s, a);
		} catch (BadLocationException e) {}
		if(pane != null){
			pane.setCaretPosition(getLength());
		}
    }
    private void setTitle(String t){
    	if(rootFrame != null){
    		if(initialTitle == null)
    			initialTitle=rootFrame.getTitle()+" "+identification;
    		rootFrame.setTitle(initialTitle+" - "+t);
    	}
    }
	public ConsoleDoc(Highlights highlights){
		this(highlights,"UTF-8");
	}
	public ConsoleDoc(Highlights highlights,String charsetName){
		this.charset=Charset.forName(charsetName);
		synchronized (globalIdentification) {
			globalIdentification++;
			identification=globalIdentification;//Get new identificaion;
		}
		textTypewrited = new StringBuffer();
        fonts= highlights;
        background= Color.black;
        foreground= Color.white;
        regularText= new SimpleAttributeSet();
        regularText.addAttributes(fonts.getRegular());
		StyleConstants.ColorConstants.setForeground(regularText, Color.white);
		StyleConstants.ColorConstants.setBackground(regularText, Color.black);        
        inputText= new SimpleAttributeSet();
        inputText.addAttributes(regularText);
		StyleConstants.ColorConstants.setForeground(inputText, Color.yellow);
		setParagraphAttributes(0, 1000, inputText, true);
	}
	public void connect(String server,int port, String key){
		final int ntry=30;
		if(pane!=null)
			pane.setVisible(true);
		String message=I18n.getString(I18n.connecting)+" ("+server+":"+port+")";
		InetAddress address;
		try {
			address= InetAddress.getByName(server);
		} catch (UnknownHostException e2) {
			setTitle(message+" "+e2.getMessage());
			return;
		}
		setTitle(message);
		for(int i=0; i<ntry;i++){
			try{
			   connection=new Socket(address,port);
			   out = connection.getOutputStream();
	           in = connection.getInputStream();
	           out.write(key.getBytes());
	           out.flush();
	           new Thread(this).start();
	           break;
	        }
			catch (Exception e) {
		        try {
		        	message+=".";
		        	setTitle(message+" "+e.getMessage());
					Thread.sleep(300);
				} catch (Exception e1) {}
				connection=null;
			}
		}
		if(connection==null){
			setTitle(I18n.getString(I18n.connectionFail));
		}
	}
	public void disconnect(){
	   if(connection!=null){
		   try {
			  connection.close();
		   } catch (IOException e) {}
	   }
	}
	public void sendControl(char c){
		try {
			out.write(c);
		} catch (Exception e) {}
	}
	public void insertString(int off, String s, javax.swing.text.AttributeSet a) throws BadLocationException{
		if(connection != null && connection.isConnected()){
			try {
				out.write(s.getBytes(charset));
				out.flush();
				synchronized (textTypewrited) {
					textTypewrited.append(s);
				}
			} catch (IOException e) {}
		}
	}
	public void remove(int off, int len) throws BadLocationException{
		if(connection != null && connection.isConnected() && off==getLength()-1 && len ==1){
			try {
				out.write(BS);
				out.flush();
			} catch (IOException e) {}
			if(pane != null)
				pane.setCaretPosition(getLength());
		}
	}
	public void run(){
		if(connection == null) return;
    	try {
			clear();
			setTitle(I18n.getString(I18n.connected));
			InputStreamReader ins= new InputStreamReader(in,charset);
			int read;
			char buf[]=new char[1000];
			while((read=ins.read(buf))!=-1){
				String s=new String(buf,0,read);
				read=s.length();
				//TODO Fix backspace functionality
				int from=0; //backspace processing 
				for(int i=0; i<read;i++){
					if(isDischargeable(s.charAt(i))){ //is visible char
						if(i>from){
							MutableAttributeSet format=regularText;
							synchronized (textTypewrited) {
								if(textTypewrited.length()>0 && s.charAt(i) == textTypewrited.charAt(0)){
									format=inputText;
									textTypewrited.deleteCharAt(0);
								}
								else{
									textTypewrited.setLength(0);
								}
							}
							append(s.substring(from, i),format);
						}
						from=i+1;
						if(s.charAt(i)==BS && getLength()>0) //Si es BackSpace
							super.remove(getLength()-1, 1);
					}
				}
				if(read>from){
					String toShow=s.substring(from, read);
					MutableAttributeSet format=regularText;
					synchronized (textTypewrited) {
						if(textTypewrited.length()>=toShow.length() &&
						   toShow.equals(textTypewrited.substring(0, toShow.length()))){
							format=inputText;
							textTypewrited.delete(0,toShow.length());
						}
						else{
							textTypewrited.setLength(0);
						}
					}
					append(toShow,format);
				}
			}
		} catch (Exception e) {}
		disconnect();
		pane.getCaret().setBlinkRate(0);
		setTitle(I18n.getString(I18n.connectionClosed));
	}
	public void setPanel(JTextPane panel) {
		this.pane = panel;
		panel.setBackground(background);
		panel.setCaretColor(foreground);
	}
	public void setRoot(java.awt.Frame raz) {
		this.rootFrame = raz;
		raz.setBackground(background);
	}
}

public class Console{
	private static final long serialVersionUID = -852166991864394680L;
	private static int offset=0; //Offset to place console at diferent location
	final ConsoleDoc inner;
	public Console(final String host, final int port, final String key, Highlights highlights, JApplet parent) {
		inner=new ConsoleDoc(highlights);
		JTextPane tp=new JTextPane();
		tp.setDocument(inner);
		inner.setPanel(tp);
		JFrame console = new JFrame();
		inner.setRoot(console);
		console.setTitle(I18n.getString(I18n.console));
		int width=highlights.getCharSize()*80+20;  //80 cols + border size
		int height=highlights.getCharHeight()*24+20; //24 lines + border size
		try{
			Point start=parent.getLocationOnScreen();
			console.setLocation(offset+start.x+(parent.getSize().width-width)/2,
					            offset+start.y+(parent.getSize().height-height)/2);
			offset = (offset+16)%160; //Update offset
		}catch (Exception e) {
		}
		console.setSize(new Dimension(width,height));
		console.setContentPane(new JScrollPane(tp));
		console.setVisible(true);
		console.toFront();
		new Thread(new Runnable(){
			public void run() {
				inner.connect(host,port,key);
			}}).start();
		tp.addKeyListener(new KeyListener(){
			public void keyPressed(KeyEvent arg) {}
			public void keyReleased(KeyEvent arg0) {}
			public void keyTyped(KeyEvent arg) {
				if(arg.isControlDown()){
					inner.sendControl(arg.getKeyChar());
				}
			}	
		});
		console.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		console.addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowClosing(java.awt.event.WindowEvent e) {
				inner.disconnect();
			}
		});
	}
}
